package example.localplugin.coolemall.recs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import schedframe.events.scheduling.SchedulingEvent;
import schedframe.exceptions.ResourceException;
import schedframe.resources.ResourceStatus;
import schedframe.resources.StandardResourceType;
import schedframe.resources.computing.Node;
import schedframe.resources.computing.ComputingResource;
import schedframe.resources.computing.Core;
import schedframe.resources.computing.profiles.energy.airthroughput.UserAirThroughputStateName;
import schedframe.resources.devices.Device;
import schedframe.resources.units.ProcessingElements;
import schedframe.resources.units.ResourceUnit;
import schedframe.resources.units.ResourceUnitName;
import schedframe.resources.units.StandardResourceUnitName;
import schedframe.scheduling.manager.resources.ClusterResourceManager;
import schedframe.scheduling.manager.resources.ResourceManager;
import schedframe.scheduling.manager.tasks.JobRegistry;
import schedframe.scheduling.plan.SchedulingPlanInterface;
import schedframe.scheduling.plan.impl.SchedulingPlan;
import schedframe.scheduling.plugin.grid.ModuleList;
import schedframe.scheduling.queue.TaskQueue;
import schedframe.scheduling.queue.TaskQueueList;
import schedframe.scheduling.tasks.TaskInterface;
import test.Node_Fan_Mapping;
import example.localplugin.BaseLocalSchedulingPlugin;
import gridsim.dcworms.DCWormsTags;

public class RECS_FCFS_Random_SP extends BaseLocalSchedulingPlugin {

	private Random rand;
	
	public RECS_FCFS_Random_SP () {
		rand = new Random(47);
		Node_Fan_Mapping.init();
	}

	public SchedulingPlanInterface<?> schedule(SchedulingEvent event, TaskQueueList queues, JobRegistry jobRegistry,
			 ResourceManager resManager, ModuleList modules) {

		ClusterResourceManager resourceManager = (ClusterResourceManager) resManager;
		SchedulingPlan plan = new SchedulingPlan();
		// choose the events types to serve.
		// Different actions for different events are possible.
		switch (event.getType()) {
		case START_TASK_EXECUTION:
		case TASK_FINISHED:
		//case TIMER:
			// our tasks are placed only in first queue (see
			// BaseLocalSchedulingPlugin.placeJobsInQueues() method)
			TaskQueue q = queues.get(0);
			
			List<Node> notSelectedNodes = null;
			try {
				notSelectedNodes = (List<Node>) resourceManager.getResourcesOfType(StandardResourceType.Node);
			} catch (ResourceException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			// check all tasks in queue
			for (int i = 0; i < q.size(); i++) {
				TaskInterface<?> task = q.get(i);
				// if status of the tasks in READY
				if (task.getStatus() == DCWormsTags.READY) {
					Map<ResourceUnitName, ResourceUnit> choosenResources = chooseResourcesForExecution(resourceManager, task);
					if (choosenResources != null) {
						addToSchedulingPlan(plan, task, choosenResources);
					}
					/*ComputingNode node = chooseRandomProvider(resourceManager);
					if (node != null) {
						//node.getAirThroughputInterface().setAirThroughputState(StandardAirThroughputStateName.FAN_ON);
						notSelectedNodes.remove(node);
						addToSchedulingPlan(plan, task, node.getName());
					}*/
				}
			}
			adjustOtherFans(notSelectedNodes);
			break;
		}
		return plan;
	}

	@SuppressWarnings("unchecked")
	private Node chooseRandomProvider(ClusterResourceManager resourceManager) {
		List<Node> nodes = null;
		try {
			nodes = (List<Node>) resourceManager.getResourcesOfType(StandardResourceType.Node);
		} catch (ResourceException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		nodes = filterNodes(nodes);
		int nodeIdx = rand.nextInt(nodes.size());
		return nodes.get(nodeIdx);
	}

	private List<Node> filterNodes(List<Node> nodes){
		List<Node> filteredNodes = new ArrayList<Node>();
		for (Node node : nodes) {
			if(node.getFreeProcessors().size() > 0)
				filteredNodes.add(node);
		}
		
		return filteredNodes;
	}
	
	private Map<ResourceUnitName, ResourceUnit> chooseResourcesForExecution(
			ClusterResourceManager resourceManager, TaskInterface<?> task) {

		Map<ResourceUnitName, ResourceUnit> map = new HashMap<ResourceUnitName, ResourceUnit>();
		
		List<Node> nodes = null;
		try {
			nodes = (List<Node>) resourceManager.getResourcesOfType(StandardResourceType.Node);
		} catch (ResourceException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		List<Node> avNodes = filterNodes(nodes, task);
		if(avNodes.size() == 0)
			return null;
		Node node = randomNode(avNodes);

		int cpuRequest;
		try {
			cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
		} catch (NoSuchFieldException e) {
			cpuRequest = 0;
		}

		if (cpuRequest != 0) {

			List<Core> cores = node.getProcessors().get(0).getCores();

			List<ComputingResource> choosenResources = new ArrayList<ComputingResource>();				
			for (int i = 0; i < cores.size(); i++) {
				if (cores.get(i).getStatus() == ResourceStatus.FREE) {
					//choosenResources.add(cores.get(i));
					cpuRequest--;
				}
			}
			if (cpuRequest > 0) {
				return null;
			}
			choosenResources.add(node.getProcessors().get(0));
			ProcessingElements pe = new ProcessingElements();
			pe.addAll(choosenResources);
			map.put(StandardResourceUnitName.PE, pe);
			return map;
		}
		return null;
	}
	
	private List<Node> filterNodes(List<Node> nodes, TaskInterface<?> task){
		List<Node> filteredNodes = new ArrayList<Node>();
		for (Node node : nodes) {
			int cpuRequest;
			try {
				cpuRequest = Double.valueOf(task.getCpuCntRequest()).intValue();
			} catch (NoSuchFieldException e) {
				cpuRequest = 0;
			}

			if (cpuRequest != 0) {

				List<Core> cores = node.getProcessors().get(0).getCores();
				if (cores.size() < cpuRequest) {
					if(cores.size() == 0){
						if(node.getProcessors().size() < cpuRequest)
							continue;
					}
				}

				int freeCores = 0;
				for(Core core: cores){
					if(core.getStatus() == ResourceStatus.FREE)
						freeCores++;
				}
				
				if(freeCores != cores.size())
					continue;
				
				filteredNodes.add(node);
			}
		}
		
		return filteredNodes;
	}
	
	private Node randomNode(List<Node> nodes){
		return nodes.get(rand.nextInt(nodes.size()));
	}
	
	private void adjustOtherFans(List<Node> nodes) {
		for (Node node : nodes) {
			
			if (node.getProcessors().size() == node.getFreeProcessorsNumber()) {
				for(Device device: node.getParent().getResourceCharacteristic().getDevices()){
					if(device.getFullName().equals(Node_Fan_Mapping.getNode_fan().get(node.getFullName()))){
						device.getAirThroughputInterface().setAirThroughputState(new UserAirThroughputStateName("Off"));
						break;
					}
				}

			} else {
				for(Device device: node.getParent().getResourceCharacteristic().getDevices()){
					if(device.getFullName().equals(Node_Fan_Mapping.getNode_fan().get(node.getFullName()))){
						device.getAirThroughputInterface().setAirThroughputState(new UserAirThroughputStateName("On"));
						break;
					}
				}
			}
		}
	}
}

